{% load i18n static %}<!DOCTYPE html>
<html>
  <head><title>{% translate 'Popup closing…' %}</title></head>
  <body>
  </script>
    <script id="popup-response-data"
            src="{% static "mizdb_tomselect/js/popup_response.js" %}"
            data-popup-response="{{ popup_response_data }}">
    </script>
  </body>